/*
 *  C64_DC.i - Dreamcast KOS specific stuff
 *
 *  Frodo (C) 1994-1997,2002 Christian Bauer
 *  Dreamcast Stuff by Tolga ABACI
 */

#include "main.h"

/*
  C64 keyboard matrix:

    Bit 7   6   5   4   3   2   1   0
  0    CUD  F5  F3  F1  F7 CLR RET DEL
  1    SHL  E   S   Z   4   A   W   3
  2     X   T   F   C   6   D   R   5
  3     V   U   H   B   8   G   Y   7
  4     N   O   K   M   0   J   I   9
  5     ,   @   :   .   -   L   P   +
  6     /   ^   =  SHR HOM  ;   *   
  7    R/S  Q   C= SPC  2  CTL  <-  1*/

#define MATRIX(a,b) (((a) << 3) | (b))

typedef struct key_seq_item
{
  int kc;
  bool shift;
} key_seq_item_t;

static key_seq_item_t d_auto_seq[4][18] =
{
  {
    { MATRIX(0, 1), false }, /*RET*/
    { MATRIX(5, 2), false }, /*L*/
    { MATRIX(4, 6), false }, /*O*/
    { MATRIX(1, 2), false }, /*A*/
    { MATRIX(2, 2), false }, /*D*/
    { MATRIX(7, 4), false }, /* */
    { MATRIX(7, 3), true  }, /*"*/
    { MATRIX(6, 1), false }, /***/
    { MATRIX(7, 3), true  }, /*"*/
    { MATRIX(5, 7), false }, /*,*/
    { MATRIX(3, 3), false }, /*8*/
    { MATRIX(5, 7), false }, /*,*/
    { MATRIX(7, 0), false }, /*1*/
    { MATRIX(0, 1), false }, /*RET*/
    { MATRIX(2, 1), false }, /*R*/
    { MATRIX(3, 6), false }, /*U*/    
    { MATRIX(4, 7), false }, /*N*/
    { MATRIX(0, 1), false }, /*RET*/
  },
  {
    { MATRIX(0, 1), false }, /*RET*/
    { MATRIX(5, 2), false }, /*L*/
    { MATRIX(4, 6), false }, /*O*/
    { MATRIX(1, 2), false }, /*A*/
    { MATRIX(2, 2), false }, /*D*/
    { MATRIX(7, 4), false }, /* */
    { MATRIX(7, 3), true  }, /*"*/
    { MATRIX(6, 1), false }, /***/
    { MATRIX(7, 3), true  }, /*"*/
    { MATRIX(5, 7), false }, /*,*/
    { MATRIX(4, 0), false }, /*9*/
    { MATRIX(5, 7), false }, /*,*/
    { MATRIX(7, 0), false }, /*1*/
    { MATRIX(0, 1), false }, /*RET*/
    { MATRIX(2, 1), false }, /*R*/
    { MATRIX(3, 6), false }, /*U*/    
    { MATRIX(4, 7), false }, /*N*/
    { MATRIX(0, 1), false }, /*RET*/
  },
  {
    { MATRIX(0, 1), false }, /*RET*/
    { MATRIX(5, 2), false }, /*L*/
    { MATRIX(4, 6), false }, /*O*/
    { MATRIX(1, 2), false }, /*A*/
    { MATRIX(2, 2), false }, /*D*/
    { MATRIX(7, 3), true  }, /*"*/
    { MATRIX(6, 1), false }, /***/
    { MATRIX(7, 3), true  }, /*"*/
    { MATRIX(5, 7), false }, /*,*/
    { MATRIX(7, 0), false }, /*1*/
    { MATRIX(4, 3), false }, /*0*/
    { MATRIX(5, 7), false }, /*,*/
    { MATRIX(7, 0), false }, /*1*/
    { MATRIX(0, 1), false }, /*RET*/
    { MATRIX(2, 1), false }, /*R*/
    { MATRIX(3, 6), false }, /*U*/    
    { MATRIX(4, 7), false }, /*N*/
    { MATRIX(0, 1), false }, /*RET*/
  },
  {
    { MATRIX(0, 1), false }, /*RET*/
    { MATRIX(5, 2), false }, /*L*/
    { MATRIX(4, 6), false }, /*O*/
    { MATRIX(1, 2), false }, /*A*/
    { MATRIX(2, 2), false }, /*D*/
    { MATRIX(7, 3), true  }, /*"*/
    { MATRIX(6, 1), false }, /***/
    { MATRIX(7, 3), true  }, /*"*/
    { MATRIX(5, 7), false }, /*,*/
    { MATRIX(7, 0), false }, /*1*/
    { MATRIX(7, 0), false }, /*1*/
    { MATRIX(5, 7), false }, /*,*/
    { MATRIX(7, 0), false }, /*1*/
    { MATRIX(0, 1), false }, /*RET*/
    { MATRIX(2, 1), false }, /*R*/
    { MATRIX(3, 6), false }, /*U*/    
    { MATRIX(4, 7), false }, /*N*/
    { MATRIX(0, 1), false }, /*RET*/
  },
};

/*
 *  Constructor, system-dependent things
 */

void C64::c64_ctor1(void)
{

}

void C64::c64_ctor2(void)
{
  gettimeofday(&tv_start, NULL); 
}


/*
 *  Destructor, system-dependent things
 */

void C64::c64_dtor(void)
{

}

/*
 *  Start main emulation thread
 */
void C64::Run(void)
{
  // Reset chips
  TheCPU->Reset();
  TheSID->Reset();
  TheCIA1->Reset();
  TheCIA2->Reset();
  TheCPU1541->Reset();

  // Patch kernal IEC routines
  orig_kernal_1d84 = Kernal[0x1d84];
  orig_kernal_1d85 = Kernal[0x1d85];
  PatchKernal(ThePrefs.FastReset, ThePrefs.Emul1541Proc);

  frame_count = 0;

  quit_thyself = false;
  autostart_index = 0;
  autostart_keytime = 5;
  thread_func();
}

/*
 *  Vertical blank: Poll keyboard and joysticks, update window
 */
void C64::VBlank(bool draw_frame)
{
  // Poll keyboard
  TheDisplay->PollKeyboard(TheCIA1->KeyMatrix, TheCIA1->RevMatrix, &joykey);

  TheCIA1->Joystick1 = 0xFF;
  TheCIA1->Joystick2 = 0xFF;

  if (ThePrefs.AutoStartDrive >= 0)
  {
    int kc = d_auto_seq[ThePrefs.AutoStartDrive][autostart_index].kc;
    bool shift = d_auto_seq[ThePrefs.AutoStartDrive][autostart_index].shift;              

    TheDisplay->FakeKeyPress(kc, shift, TheCIA1->KeyMatrix, TheCIA1->RevMatrix, &joykey);

    autostart_keytime --;
    if (autostart_keytime == 0)
    {
      autostart_keytime = 1;
      autostart_index ++;
      
      if (autostart_index == 18)
      {       
	ThePrefs.AutoStartDrive = -1;
	autostart_index = 0;
	autostart_keytime = 5;
      }
    }
  }
  else if (!TheDisplay->HandleUserInput())
  {
    // Poll joysticks
    TheCIA1->Joystick1 = poll_joystick(0);
    TheCIA1->Joystick2 = poll_joystick(1);
  }

  if (ThePrefs.JoystickSwap) {
    uint8 tmp = TheCIA1->Joystick1;
    TheCIA1->Joystick1 = TheCIA1->Joystick2;
    TheCIA1->Joystick2 = tmp;
  }

  // Joystick keyboard emulation
  if (TheDisplay->NumLock())
    TheCIA1->Joystick1 &= joykey;
  else
    TheCIA1->Joystick2 &= joykey;
       
  // Count TOD clocks
  TheCIA1->CountTOD();
  TheCIA2->CountTOD();

  // KOS/PVR demand that we miss no frames at all...

  // Update window if needed 
  if (draw_frame) 
  {
    // Calculate time between VBlanks, display speedometer
    struct timeval tv;
    gettimeofday(&tv, NULL);
    if ((tv.tv_usec -= tv_start.tv_usec) < 0) {
      tv.tv_usec += 1000000;
      tv.tv_sec -= 1;
    }
    tv.tv_sec -= tv_start.tv_sec;
    double elapsed_time = (double)tv.tv_sec * 1000000 + tv.tv_usec;
    speed_index = 20000 / (elapsed_time + 1) * ThePrefs.SkipFrames * 100;

    if (speed_index < 49)
    {
      // Wow we're really slow.
      TheDisplay->UpdatePVR();
      TheDisplay->UpdatePVR();
      TheDisplay->UpdatePVR();
      TheDisplay->UpdatePVR();
      speed_index = 25;
      frame_count += 4;
    }
    else if (speed_index < 99) 
    {      
      TheDisplay->UpdatePVR();
      TheDisplay->UpdatePVR();
      speed_index = 50;
      frame_count += 2;
    }
    else 
    {
      TheDisplay->UpdatePVR();
      speed_index = 100;
      frame_count ++;
    }
    
    if (!ThePrefs.PALMode)
      while (frame_count >= 5)
      {
	frame_count -= 5;
	TheDisplay->UpdatePVR();
      }

    gettimeofday(&tv_start, NULL);    

    TheDisplay->Speedometer((int) speed_index);
    TheDisplay->Update();
  }
  else 
  {
    TheDisplay->UpdatePVR();
    frame_count ++;
  }
}

void C64::QuitThyself()
{
  quit_thyself = true;
}

void C64::open_close_joysticks(bool oldjoy1, bool oldjoy2, bool newjoy1, bool newjoy2)
{   
}

uint8 C64::poll_joystick(int port)
{
  // Get joystick status from Maple
  maple_device_t *cont_dev = maple_enum_type(port, MAPLE_FUNC_CONTROLLER);
  if (cont_dev == NULL)
    return 0xff;   // No Joystick in port
  cont_state_t *cont_state = (cont_state_t*) maple_dev_status(cont_dev);

  uint8 j = 0xff;

  if (cont_state->buttons & CONT_DPAD_LEFT)
    j &= 0xfb;							// Left
  if (cont_state->buttons & CONT_DPAD_RIGHT)
    j &= 0xf7;							// Right
  if (cont_state->buttons & CONT_DPAD_UP)
    j &= 0xfe;							// Up
  if (cont_state->buttons & CONT_DPAD_DOWN)
    j &= 0xfd;							// Down
  if (cont_state->buttons & CONT_A)
    j &= 0xef;							// Button

  if (cont_state->joyx < -40)
    j &= 0xfb;							// Left
  else if (cont_state->joyx > 40)
    j &= 0xf7;							// Right
  
  if (cont_state->joyy < -40)
    j &= 0xfe;							// Up
  else if (cont_state->joyy > 40)
    j &= 0xfd;							// Down

  return j;
}

/*
 * The emulation's main loop
 */

void C64::thread_func(void)
{
	int linecnt = 0;

#ifdef FRODO_SC
	while (!quit_thyself) {

		// The order of calls is important here
		if (TheVIC->EmulateCycle())
			TheSID->EmulateLine();
		TheCIA1->CheckIRQs();
		TheCIA2->CheckIRQs();
		TheCIA1->EmulateCycle();
		TheCIA2->EmulateCycle();
		TheCPU->EmulateCycle();

		if (ThePrefs.Emul1541Proc) {
			TheCPU1541->CountVIATimers(1);
			if (!TheCPU1541->Idle)
				TheCPU1541->EmulateCycle();
		}
		CycleCounter++;
#else
	while (!quit_thyself) {

		// The order of calls is important here
		int cycles = TheVIC->EmulateLine();

     		TheSID->EmulateLine();
#if !PRECISE_CIA_CYCLES
		TheCIA1->EmulateLine(ThePrefs.CIACycles);
		TheCIA2->EmulateLine(ThePrefs.CIACycles);
#endif

		if (ThePrefs.Emul1541Proc) {
			int cycles_1541 = ThePrefs.FloppyCycles;
			TheCPU1541->CountVIATimers(cycles_1541);

			if (!TheCPU1541->Idle) {
				// 1541 processor active, alternately execute
				//  6502 and 6510 instructions until both have
				//  used up their cycles
				while (cycles >= 0 || cycles_1541 >= 0)
					if (cycles > cycles_1541)
						cycles -= TheCPU->EmulateLine(1);
					else
						cycles_1541 -= TheCPU1541->EmulateLine(1);
			} else
				TheCPU->EmulateLine(cycles);
		} else
			// 1541 processor disabled, only emulate 6510
			TheCPU->EmulateLine(cycles);
#endif
		linecnt++;

	}
}

void C64::Pause()
{
  if (!have_a_break)
  {
    have_a_break = true;
    TheSID->PauseSound();
  }
}

void C64::Resume()
{
  if (have_a_break)
  {
    have_a_break = false;
    TheSID->ResumeSound();
  }
}
