/* 
 * Dreamcast GUI Functions
 *
 * by Tolga ABACI, 2003
*/

#include "DCGUI.h"

/* Creates menu with default settings */
void menu_create(menu_t *mn, float x, float y, float w, float h, float z, menu_item_t *items, int cnt, plx_font_t* fnt, plx_fcxt_t* fnt_cxt)
{
  /* Start blank */
  memset(mn, 0, sizeof(*mn));
  
  /* First colors */
  mn->col.border.r = mn->col.border.g = mn->col.border.b = 0.8f;
  mn->col.icon.r = mn->col.icon.g = mn->col.icon.b = 1.0f;
  mn->col.in.r = mn->col.in.b = 0.1f;
  mn->col.in.g = 0.5;
  mn->col.sel.r = mn->col.sel.g = 0.4f;
  mn->col.sel.b = 0.4f;

  /* Then items */
  mn->items.list = items;
  mn->items.cnt = cnt;
  mn->items.fnt = fnt;
  mn->items.fnt_cxt = fnt_cxt;
  mn->items.sel = 0;
  mn->items.start = 0;
 
  float down, up;
  plx_fcxt_str_metrics(mn->items.fnt_cxt, "Tg",
		       0, &up, 0, &down);
 
  /* Then dimensions */ 
  mn->dim.x = x;
  mn->dim.y = y;
  mn->dim.w = w;
  mn->dim.h = h;  
  mn->dim.z = z;  
  mn->dim.lh = 2.0f;
  mn->dim.b = 2.0f;  
  mn->dim.mw = 10.0f;    
  mn->dim.mh = 10.0f;    

  mn->items.max = (int) ((mn->dim.h - mn->dim.mh) / (up + down + 2 * mn->dim.lh));

  /* Then the state/options */
  mn->stat.sub_open = 0;
  mn->opt.dim_text = 0;
  mn->opt.scroll = (cnt > mn->items.max);
  mn->opt.scroll_sm = mn->opt.scroll;
}

void tick_draw_pvr(float x, float y, float w, float h, float z, float a, float r, float g, float b)
{
  pvr_vertex_t vert;

  vert.flags = PVR_CMD_VERTEX;
  vert.x = x + w / 3;
  vert.y = y + h;
  vert.z = z;
  vert.u = 0.0f;
  vert.v = 0.0f;
  vert.argb = PVR_PACK_COLOR(a, r, g, b);
  vert.oargb = 0;
  pvr_prim(&vert, sizeof(vert));

  vert.x = x + w / 3;
  vert.y = y + 3 * h / 4;
  pvr_prim(&vert, sizeof(vert));  

  vert.flags = PVR_CMD_VERTEX_EOL;
  vert.x = x + w;
  vert.y = y;
  pvr_prim(&vert, sizeof(vert));

  vert.flags = PVR_CMD_VERTEX;
  vert.x = x;
  vert.y = y + h / 2;
  pvr_prim(&vert, sizeof(vert));

  vert.x = x + w / 3;
  vert.y = y + 3 * h / 4;
  pvr_prim(&vert, sizeof(vert));  

  vert.flags = PVR_CMD_VERTEX_EOL;
  vert.x = x + w / 3;
  vert.y = y + h;
  pvr_prim(&vert, sizeof(vert));
}

void smarrow_draw_pvr(float x, float y, float w, float h, float z, float a, float r, float g, float b)
{
  pvr_vertex_t vert;

  vert.flags = PVR_CMD_VERTEX;
  vert.x = x + w;
  vert.y = y + h / 2;
  vert.z = z;
  vert.u = 0.0f;
  vert.v = 0.0f;
  vert.argb = PVR_PACK_COLOR(a, r, g, b);
  vert.oargb = 0;
  pvr_prim(&vert, sizeof(vert));

  vert.x = x;
  vert.y = y + h;
  pvr_prim(&vert, sizeof(vert));  

  vert.flags = PVR_CMD_VERTEX_EOL;
  vert.x = x;
  vert.y = y;
  pvr_prim(&vert, sizeof(vert));
}

void rect_draw_pvr(float x1, float y1, float x2, float y2, float z, 
		   float a, float r, float g, float b)
{
  pvr_vertex_t vert;

  vert.flags = PVR_CMD_VERTEX;
  vert.x = x1;
  vert.y = y2;
  vert.z = z;
  vert.u = 0.0f;
  vert.v = 0.0f;
  vert.argb = PVR_PACK_COLOR(a, r, g, b);
  vert.oargb = 0;
  pvr_prim(&vert, sizeof(vert));
  
  vert.y = y1;
  pvr_prim(&vert, sizeof(vert));

  vert.x = x2;
  vert.y = y2;
  pvr_prim(&vert, sizeof(vert));
  
  vert.flags = PVR_CMD_VERTEX_EOL;
  vert.y = y1;
  pvr_prim(&vert, sizeof(vert));
}

void box_draw_pvr(float x, float y, float w, float h, float z, float b,
		  float ia, float ir, float ig, float ib, 
		  float ba, float br, float bg, float bb)
{
  rect_draw_pvr(x, y, x + w, y + h, z, ba, br, bg, bb);
  rect_draw_pvr(x + b, y + b, x + w - b, y + h - b, z + 0.01f, ia, ir, ig, ib);
}

void menu_draw_box_pvr(menu_t *mn)
{
  pvr_poly_cxt_t cxt;
  pvr_poly_hdr_t hdr;

  pvr_poly_cxt_col(&cxt, PVR_LIST_OP_POLY);
  cxt.gen.shading = PVR_SHADE_FLAT;
  pvr_poly_compile(&hdr, &cxt);
  pvr_prim(&hdr, sizeof(hdr));

  if (mn->stat.sub_open)
  {
    mn->col.border.r /= 2.0f;
    mn->col.border.g /= 2.0f;
    mn->col.border.b /= 2.0f;

    mn->col.icon.r /= 2.0f;
    mn->col.icon.g /= 2.0f;
    mn->col.icon.b /= 2.0f;

    mn->col.in.r /= 2.0f;
    mn->col.in.g /= 2.0f;
    mn->col.in.b /= 2.0f;

    mn->col.sel.r /= 2.0f;
    mn->col.sel.g /= 2.0f;
    mn->col.sel.b /= 2.0f;
  }

  rect_draw_pvr(mn->dim.x, mn->dim.y, 
		mn->dim.x + mn->dim.w, mn->dim.y + mn->dim.h, 
		mn->dim.z,		
		1.0f, mn->col.border.r, mn->col.border.g, mn->col.border.b);
  rect_draw_pvr(mn->dim.x + mn->dim.b, mn->dim.y + mn->dim.b, 
		mn->dim.x + mn->dim.w - mn->dim.b, mn->dim.y + mn->dim.h - mn->dim.b, 
		mn->dim.z + 0.01f,		
		1.0f, mn->col.in.r, mn->col.in.g, mn->col.in.b);

  float up, down, y;
  plx_fcxt_str_metrics(mn->items.fnt_cxt, "Tg",
		       0, &up, 0, &down);

  if (mn->items.sel >= 0 && !mn->stat.sub_open)
  {
    y = mn->dim.mh + (up + down + 2 * mn->dim.lh) * (mn->items.sel - mn->items.start); 

    rect_draw_pvr(mn->dim.x + (mn->dim.mw / 2), mn->dim.y + y - mn->dim.lh,
		  mn->dim.x + mn->dim.w - (mn->dim.mw / 2), mn->dim.y + y + up + down + mn->dim.lh, 
		  mn->dim.z + 0.02f,		
		  1.0f, mn->col.sel.r, mn->col.sel.g, mn->col.sel.b);
  }

  y = mn->dim.mh;

  int end = mn->items.start + mn->items.max;
  if (end > mn->items.cnt)
    end = mn->items.cnt;

  for (int i = mn->items.start; i < end; i ++)
  {   
    if (mn->items.list[i].icon == ICON_TICK)
      tick_draw_pvr(mn->dim.x + mn->dim.w - mn->dim.mw - (up + down), mn->dim.y + y, 
		    up + down, up + down + 2 * mn->dim.lh, 
		    mn->dim.z + 0.03f,
		    1.0f, mn->col.icon.r, mn->col.icon.g, mn->col.icon.b); 
    else if (mn->items.list[i].icon == ICON_SMARROW)
      smarrow_draw_pvr(mn->dim.x + mn->dim.w - mn->dim.mw - (up + down) / 2, mn->dim.y + y + mn->dim.lh, 
		       (up + down) / 2, up, 
		       mn->dim.z + 0.03f,
		       1.0f, mn->col.icon.r, mn->col.icon.g, mn->col.icon.b); 

    y += up + down + 2 * mn->dim.lh;
  }

  if (mn->stat.sub_open)
  {
    mn->col.border.r *= 2.0f;
    mn->col.border.g *= 2.0f;
    mn->col.border.b *= 2.0f;

    mn->col.icon.r *= 2.0f;
    mn->col.icon.g *= 2.0f;
    mn->col.icon.b *= 2.0f;

    mn->col.in.r *= 2.0f;
    mn->col.in.g *= 2.0f;
    mn->col.in.b *= 2.0f;

    mn->col.sel.r *= 2.0f;
    mn->col.sel.g *= 2.0f;
    mn->col.sel.b *= 2.0f;
  }

}

void menu_auto_sub_menu_pos(menu_t *mn)
{
  float up, down;
  plx_fcxt_str_metrics(mn->items.fnt_cxt, "Tg",
		       0, &up, 0, &down);

  for (int i = 0; i < mn->items.cnt; i ++)
  {
    float y;
    menu_t *sub_menu; 
    sub_menu = mn->items.list[i].sub_menu;	

    if (sub_menu)
    {
      y = mn->dim.mh + (up + down + 2 * mn->dim.lh) * i; 

      sub_menu->dim.x = mn->dim.x + mn->dim.w - mn->dim.mw;
      sub_menu->dim.y = mn->dim.y + y - sub_menu->dim.mh + (mn->dim.mh / 2);
      sub_menu->dim.z = mn->dim.z + 1.0f;

      menu_auto_sub_menu_pos(sub_menu);
    }
  }
}

void menu_draw_box_chain_pvr(menu_t *mn)
{
  menu_draw_box_pvr(mn);
  if (mn->stat.sub_open)
    if (mn->items.sel != -1)
    {
      menu_t *sub_menu; 
      
      if ((sub_menu = mn->items.list[mn->items.sel].sub_menu))
      {       
	if (mn->opt.scroll_sm)
	{
	  float up, down, y;

	  plx_fcxt_str_metrics(mn->items.fnt_cxt, "Tg",
			       0, &up, 0, &down);

	  y = mn->dim.mh + (up + down + 2 * mn->dim.lh) * (mn->items.sel - mn->items.start); 
	  
	  sub_menu->dim.x = mn->dim.x + mn->dim.w - mn->dim.mw;
	  sub_menu->dim.y = mn->dim.y + y - sub_menu->dim.mh;
	  sub_menu->dim.z = mn->dim.z + 1.0f;
	}
		  
	menu_draw_box_chain_pvr(mn->items.list[mn->items.sel].sub_menu);
      }
    }
}

void menu_draw_text_pvr(menu_t *mn)
{
  float up, down, y;
  plx_fcxt_str_metrics(mn->items.fnt_cxt, "Tg",
		       0, &up, 0, &down);

  y = mn->dim.mh;

  uint32 color = mn->items.fnt_cxt->color;
  float fa, fr, fg, fb;

  fa = ((color >> 24) & 0xFF) / 255.0f;
  fr = ((color >> 16) & 0xFF) / 255.0f;
  fg = ((color >> 8) & 0xFF) / 255.0f;
  fb = ((color >> 0) & 0xFF) / 255.0f;

  if (mn->stat.sub_open || mn->opt.dim_text)
    mn->items.fnt_cxt->color = PVR_PACK_COLOR(fa, fr / 2.0f, fg / 2.0f, fb / 2.0f);

  int end = mn->items.start + mn->items.max;
  if (end > mn->items.cnt)
    end = mn->items.cnt;

  for (int i = mn->items.start; i < end; i ++)
  {
    if (mn->stat.sub_open && (mn->items.sel == i))
      mn->items.fnt_cxt->slant = mn->items.fnt_cxt->size / 5.0f;

    plx_fcxt_begin(mn->items.fnt_cxt);

    y += up + mn->dim.lh;    
    plx_fcxt_setpos(mn->items.fnt_cxt, 
		    mn->dim.x + mn->dim.mw, mn->dim.y + y, mn->dim.z + 0.03f);

    plx_fcxt_draw(mn->items.fnt_cxt, mn->items.list[i].text);
    y += down + mn->dim.lh;

    plx_fcxt_end(mn->items.fnt_cxt);

    if (mn->stat.sub_open && (mn->items.sel == i))
      mn->items.fnt_cxt->slant = 0.0f;
  }

  if (mn->stat.sub_open || mn->opt.dim_text)
    mn->items.fnt_cxt->color = color;
}

void menu_draw_text_chain_pvr(menu_t *mn)
{
  menu_draw_text_pvr(mn);
  if (mn->stat.sub_open)
    if (mn->items.sel != -1)
      if (mn->items.list[mn->items.sel].sub_menu)
	menu_draw_text_chain_pvr(mn->items.list[mn->items.sel].sub_menu);
}

int menu_operate(menu_t *mn, cont_state_t *cont_state)
{
  /* Find out active menu */
  menu_t *act_mn, *act_mn_p;
  act_mn_p = 0;
  act_mn = mn;
  while (act_mn->stat.sub_open) 
  {
    act_mn_p = act_mn;
    act_mn = act_mn->items.list[act_mn->items.sel].sub_menu;
  }

  if (act_mn->items.sel == -1)
  {
    if (cont_state->buttons & CONT_B)
      return MENU_CLOSE;
    else 
      return MENU_NOACTION;
  }

  if ((cont_state->buttons & CONT_DPAD_RIGHT) || (cont_state->buttons & CONT_A))
  {
    if (act_mn->items.list[act_mn->items.sel].sub_menu)
      act_mn->stat.sub_open = 1;
    else if (cont_state->buttons & CONT_A)
      return act_mn->items.list[act_mn->items.sel].id;
  }
  else if ((cont_state->buttons & CONT_DPAD_LEFT) || (cont_state->buttons & CONT_B))  
  {
    if (act_mn_p)
    {
      act_mn_p->stat.sub_open = 0;
      act_mn->items.sel = 0;
      act_mn->items.start = 0;
    }
    else 
      return MENU_CLOSE;
  }
  else if (cont_state->buttons & CONT_DPAD_UP)
  {
    act_mn->items.sel --;
    if (act_mn->items.sel < 0)
    {
      if (act_mn->opt.scroll)
	act_mn->items.sel = 0;
      else 
	act_mn->items.sel = act_mn->items.cnt - 1;      
    }

    if (act_mn->items.sel < act_mn->items.start)
      act_mn->items.start --;
    
  }
  else if (cont_state->buttons & CONT_DPAD_DOWN)
  {
    act_mn->items.sel ++;
    if (act_mn->items.sel >= act_mn->items.cnt)
    {
      if (act_mn->opt.scroll)
	act_mn->items.sel = act_mn->items.cnt - 1;
      else 
	act_mn->items.sel = 0;
    }

    if (act_mn->items.sel >= act_mn->items.start + act_mn->items.max)
      act_mn->items.start ++;
  }
  else if (cont_state->joyy < -40 && act_mn->opt.scroll)
  {
    int amt = (-cont_state->joyy - 30) / 10;

    act_mn->items.sel -= amt;
    if (act_mn->items.sel < 0)
      act_mn->items.sel = 0;
    if (act_mn->items.sel < act_mn->items.start)      
      act_mn->items.start = act_mn->items.sel;
  }
  else if (cont_state->joyy > 40 && act_mn->opt.scroll)
  {
    int amt = (cont_state->joyy - 30) / 10;

    act_mn->items.sel += amt;
    if (act_mn->items.sel >= act_mn->items.cnt)
      act_mn->items.sel = act_mn->items.cnt - 1;
    if (act_mn->items.sel >= act_mn->items.start + act_mn->items.max)      
      act_mn->items.start = act_mn->items.sel - act_mn->items.max + 1;
  } 
  
  return MENU_NOACTION;
}

void fileselector_create(fileselector_t *fls, menu_t *menu, char *dir)
{
  int cap = 10;
  int cnt = 0;

  if (dir)
    fls->dir = dir;
  else 
    dir = fls->dir;

  char *flnv = (char*) malloc(cap * MAX_FN_LEN * sizeof(char));
  
  file_t dir_hnd = fs_open(dir, O_DIR | O_RDONLY);
  while (dirent_t *direntry = fs_readdir(dir_hnd))
  {
    if (strcmp(direntry->name, ".") && strcmp(direntry->name, ".."))
    {
      cnt ++;
      if (cnt > cap)
      {
	char *flnv_new = (char *) malloc((cap + 25) * MAX_FN_LEN * sizeof(char));
	memcpy(flnv_new, flnv, cap * MAX_FN_LEN * sizeof(char));
	free(flnv);
	flnv = flnv_new;
	cap += 25;      
      }
   
      memcpy(flnv + (cnt - 1) * MAX_FN_LEN, direntry->name, MAX_FN_LEN);     
    }
  }
  fs_close(dir_hnd);  

  fls->menu = menu;
  fls->fln.vector = flnv;
  fls->fln.cnt = cnt;
  fls->follow_dirs = 1;

  menu_item_t *list;
  if (cnt != 0) 
  {
    list = (menu_item_t*) malloc(cnt * sizeof(menu_item_t)); 
  
    for (int i = 0; i < cnt; i ++)
    {
      list[i].text = flnv + i * MAX_FN_LEN;
      list[i].id = (int) (flnv + i * MAX_FN_LEN);
      list[i].icon = ICON_NONE;
      list[i].sub_menu = 0;
    }

    menu->items.sel = 0;  
    menu->opt.dim_text = 0;    
  }
  else 
  {
    list = (menu_item_t*) malloc(sizeof(menu_item_t)); 

    list[0].text = "-Empty-";;
    list[0].id = 0;
    list[0].icon = ICON_NONE;
    list[0].sub_menu = 0;
    
    menu->items.sel = -1;  
    menu->opt.dim_text = 1;    
    cnt = 1;
  }

  menu->items.list = list;
  menu->items.cnt = cnt;
  menu->items.start = 0;    

  menu->opt.scroll = 1;
  menu->opt.scroll_sm = 1;
}

void fileselector_destroy(fileselector_t *fls)
{
  free(fls->fln.vector);
  free(fls->menu->items.list);
  fls->menu->items.cnt = 0;
}

int fileselector_operate(fileselector_t *fls, cont_state_t *cont_state, char **fln_out)
{
  int id = menu_operate(fls->menu, cont_state);

  if (id == MENU_CLOSE)
    return FLS_CLOSE;
  else if (id == MENU_NOACTION && !(cont_state->buttons & CONT_Y && fls->follow_dirs))
    return FLS_NOACTION;
  else
  {
    char *fln, *full_fln; 

    fln = (char*) id;               

    if (cont_state->buttons & CONT_Y)
    {
      int i = strlen(fls->dir) - 1; 
      while (i >= 0 && fls->dir[i] != '/')
	i --;
      if (i == 0)
	i = 1;
      fls->dir[i] = 0;
      full_fln = fls->dir;
    }
    else 
    {
      full_fln = (char *) malloc((strlen(fls->dir) + strlen(fln) + 2) * sizeof(char));
      if (!strcmp(fls->dir, "/"))
	sprintf(full_fln, "/%s", fln);
      else
	sprintf(full_fln, "%s/%s", fls->dir, fln);
    }

    file_t hnd = fs_open(full_fln, O_DIR | O_RDONLY);
    if (hnd && fls->follow_dirs)
    {
      fs_close(hnd);      
      fileselector_destroy(fls);
      if (full_fln != fls->dir)
	free(fls->dir);
      fileselector_create(fls, fls->menu, full_fln);
    }
    else 
    {
      *fln_out = full_fln;
      return FLS_OK;
    }
  }
  
  return FLS_NOACTION;
}

void requester_create(requester_t *req, float x, float y, float w, float h, float cw, float ch, float z, 
		      char *msg, menu_item_t *choice_items, int choice_cnt, 
		      plx_font_t* fnt, plx_fcxt_t* fnt_cxt)
{
  req->msg_menu_item.text = msg;
  req->msg_menu_item.id = 0;
  req->msg_menu_item.icon = ICON_NONE;
  req->msg_menu_item.sub_menu = 0;

  menu_create(&(req->msg_menu), x, y, w, h, z, &(req->msg_menu_item), 1, fnt, fnt_cxt);
  req->msg_menu.dim.mw = 0.1f * w;
  req->msg_menu.dim.mh = 0.4f * h;

  menu_create(&(req->choice_menu), x + w - cw, y + h + 10.0f, cw, ch, z, choice_items, choice_cnt, fnt, fnt_cxt);
}

void controller_create(controller_t *ctlr, int port, int intval)
{
  ctlr->cont.port = port;
  ctlr->rpt.intval = intval;
  ctlr->rpt.t_left = intval;

  maple_device_t *cont_dev = maple_enum_type(port, MAPLE_FUNC_CONTROLLER);
  if (cont_dev != NULL)
  {
    cont_state_t *cont_state = (cont_state_t*) maple_dev_status(cont_dev);
    ctlr->cont.last_st = *cont_state;    
  }
}

bool controller_check_input(controller_t *ctlr)
{
  maple_device_t *cont_dev = maple_enum_type(ctlr->cont.port, MAPLE_FUNC_CONTROLLER);
  if (cont_dev != NULL)
  {
    cont_state_t *cont_state = (cont_state_t*) maple_dev_status(cont_dev);

    if (ctlr->rpt.t_left == 0 ||
	cont_state->buttons != ctlr->cont.last_st.buttons ||
	cont_state->ltrig != ctlr->cont.last_st.ltrig ||
	cont_state->rtrig != ctlr->cont.last_st.rtrig ||
	cont_state->joyx != ctlr->cont.last_st.joyx ||
	cont_state->joyy != ctlr->cont.last_st.joyy ||
	cont_state->joy2x != ctlr->cont.last_st.joy2x ||
	cont_state->joy2y != ctlr->cont.last_st.joy2y)
    {
      ctlr->cont.last_st = *cont_state;    
      ctlr->rpt.t_left = ctlr->rpt.intval;
      return true;
    }
    else
    {
      ctlr->rpt.t_left --;
      return false;
    }
  }
  else
    return false;
}

void keyboard_draw_box_pvr(keyboard_t *keybd)
{
  pvr_poly_cxt_t cxt;
  pvr_poly_hdr_t hdr;

  pvr_poly_cxt_col(&cxt, PVR_LIST_OP_POLY);
  cxt.gen.shading = PVR_SHADE_FLAT;
  pvr_poly_compile(&hdr, &cxt);
  pvr_prim(&hdr, sizeof(hdr));

  box_draw_pvr(keybd->dim.x, keybd->dim.y, keybd->dim.w, keybd->dim.h, keybd->dim.z, keybd->dim.b,
	       1.0f, keybd->col.in.r, keybd->col.in.g, keybd->col.in.b,
	       1.0f, keybd->col.border.r, keybd->col.border.g, keybd->col.border.b);

  for (int i = 0; i < keybd->row.cnt; i ++)
  {
    key_row_t *cur_row = &(keybd->row.list[i]);
    float x = cur_row->x + keybd->dim.x;
    float y = keybd->dim.y + keybd->row.list[i].y;
    
    for (int j = 0; j < cur_row->key.cnt; j ++)
    {
      if (cur_row == keybd->sel.row && j == keybd->sel.index) 
	box_draw_pvr(x, y, cur_row->key.list[j].w, cur_row->h, keybd->dim.z + 1.0f, keybd->dim.b,
		     1.0f, keybd->col.sel.r, keybd->col.sel.g, keybd->col.sel.b,
		     1.0f, keybd->col.border.r, keybd->col.border.g, keybd->col.border.b);
      else
	box_draw_pvr(x, y, cur_row->key.list[j].w, cur_row->h, keybd->dim.z + 1.0f, keybd->dim.b,
		     1.0f, keybd->col.key.r, keybd->col.key.g, keybd->col.key.b,
		     1.0f, keybd->col.border.r, keybd->col.border.g, keybd->col.border.b);
    
      x += cur_row->key.list[j].w + cur_row->cw;
    }
  }
}

void keyboard_draw_text_pvr(keyboard_t *keybd)
{
  pvr_poly_cxt_t cxt;
  pvr_poly_hdr_t hdr;

  pvr_poly_cxt_col(&cxt, PVR_LIST_TR_POLY);
  cxt.gen.shading = PVR_SHADE_FLAT;
  pvr_poly_compile(&hdr, &cxt);
  pvr_prim(&hdr, sizeof(hdr));

  for (int i = 0; i < keybd->row.cnt; i ++)
  {
    key_row_t *cur_row = &(keybd->row.list[i]);
    float x = cur_row->x + keybd->dim.x;
    float y = keybd->dim.y + keybd->row.list[i].y;
    
    for (int j = 0; j < cur_row->key.cnt; j ++)
    {
      char *top = cur_row->key.list[j].text.top;
      char *bot = cur_row->key.list[j].text.bot;

      float t_left, t_right, t_up, t_down;
      float b_left, b_right, b_up, b_down;
      float t_x, t_y, b_x, b_y;

      if (top)
      {
	float size = keybd->txt.fnt_cxt->size;
	keybd->txt.fnt_cxt->size *= 0.70;

	plx_fcxt_str_metrics(keybd->txt.fnt_cxt, bot,
			     &b_left, &b_up, &b_right, &b_down);
	plx_fcxt_str_metrics(keybd->txt.fnt_cxt, top,
			   &t_left, &t_up, &t_right, &t_down);    

	b_x = x + (cur_row->key.list[j].w - b_left - b_right) / 2.0f + b_left;
	b_y = y + (cur_row->h * 0.6f - b_up - b_down) / 2.0f + cur_row->h * 0.4f + b_up;

	t_x = x + (cur_row->key.list[j].w - t_left - t_right) / 2.0f + t_left;	
	t_y = y + (cur_row->h * 0.6f - t_up - t_down) / 2.0f + t_up;

	plx_fcxt_begin(keybd->txt.fnt_cxt);
	plx_fcxt_setpos(keybd->txt.fnt_cxt, 
			t_x, t_y, keybd->dim.z + 2.0f);
	plx_fcxt_draw(keybd->txt.fnt_cxt, top);
	plx_fcxt_end(keybd->txt.fnt_cxt);

	plx_fcxt_begin(keybd->txt.fnt_cxt);
	plx_fcxt_setpos(keybd->txt.fnt_cxt, 
			b_x, b_y, keybd->dim.z + 2.0f);
	plx_fcxt_draw(keybd->txt.fnt_cxt, bot);
	plx_fcxt_end(keybd->txt.fnt_cxt);

	keybd->txt.fnt_cxt->size = size;
      }
      else
      {
	plx_fcxt_str_metrics(keybd->txt.fnt_cxt, bot,
			     &b_left, &b_up, &b_right, &b_down);

	b_x = x + (cur_row->key.list[j].w - b_left - b_right) / 2.0f + b_left;

	b_y = y + (cur_row->h - b_up - b_down) / 2.0f + b_up;     
      
	plx_fcxt_begin(keybd->txt.fnt_cxt);
	plx_fcxt_setpos(keybd->txt.fnt_cxt, 
			b_x, b_y, keybd->dim.z + 2.0f);
	plx_fcxt_draw(keybd->txt.fnt_cxt, bot);
	plx_fcxt_end(keybd->txt.fnt_cxt);
      }
    
      x += cur_row->key.list[j].w + cur_row->cw;
    }
  }
}

void keyboard_create(keyboard_t *keybd, float x, float y, float w, float h, float z, 
		     plx_font_t* fnt, plx_fcxt_t* fcxt, key_row_t *row_list, int row_cnt)
{
  keybd->dim.x = x;
  keybd->dim.y = y;
  keybd->dim.w = w;
  keybd->dim.h = h;
  keybd->dim.z = z;
  keybd->dim.b = 2.0f;

  keybd->sel.row = &(row_list[0]);
  keybd->sel.index = 0;

  keybd->txt.fnt = fnt;
  keybd->txt.fnt_cxt = fcxt;

  keybd->row.list = row_list;
  keybd->row.cnt = row_cnt;

  keybd->col.border.r = keybd->col.border.g = keybd->col.border.b = 0.8f;
  keybd->col.in.r = keybd->col.in.g = 0.2f;
  keybd->col.in.b = 0.2f;
  keybd->col.sel.r = keybd->col.sel.g = keybd->col.sel.b = 0.5f;
  keybd->col.key.r = keybd->col.key.g = keybd->col.key.b = 0.3f;
}

int keyboard_operate(keyboard_t *keybd, cont_state_t *cont_state)
{
  if (cont_state->buttons & CONT_DPAD_RIGHT)
  {
    keybd->sel.index ++;

    if (keybd->sel.index >= keybd->sel.row->key.cnt)
    {
      if (keybd->sel.row->dir.right)
      {
	keybd->sel.row = keybd->sel.row->dir.right;
	keybd->sel.index = 0;
      }
      else 
	keybd->sel.index --;
    }
  }
  else if (cont_state->buttons & CONT_DPAD_LEFT)
  {
    keybd->sel.index --;

    if (keybd->sel.index < 0)
    {
      if (keybd->sel.row->dir.left)
      {
	keybd->sel.row = keybd->sel.row->dir.left;
	keybd->sel.index = keybd->sel.row->key.cnt - 1;
      }
      else 
	keybd->sel.index = 0;
    }      
  }
  else if (cont_state->buttons & CONT_DPAD_DOWN)
  {
    if (keybd->sel.row->dir.down)
    {
      keybd->sel.row = keybd->sel.row->dir.down;
      if (keybd->sel.index >= keybd->sel.row->key.cnt)
	keybd->sel.index = keybd->sel.row->key.cnt - 1;
    }
  }
  else if (cont_state->buttons & CONT_DPAD_UP)
  {
    if (keybd->sel.row->dir.up)
    {
      keybd->sel.row = keybd->sel.row->dir.up;
      if (keybd->sel.index >= keybd->sel.row->key.cnt)
	keybd->sel.index = keybd->sel.row->key.cnt - 1;
    }
  }
  else if (cont_state->buttons & CONT_A)
    return keybd->sel.row->key.list[keybd->sel.index].id;
    
  return KEY_NOACTION;
}
