/* 
 * Dreamcast GUI Functions
 *
 * by Tolga ABACI, 2003
*/

#ifndef DCGUI_H
#define DCGUI_H

#include <kos.h>
#include <time.h>
#include <stdlib.h>

#include <plx/font.h>
#include <plx/sprite.h>
#include <plx/list.h>
#include <plx/dr.h>

#define KEY_NOACTION    -1

#define MENU_NOACTION   -1
#define MENU_CLOSE      -2

#define FLS_NOACTION    -1
#define FLS_CLOSE       -2
#define FLS_OK          -3

#define ICON_NONE       0
#define ICON_TICK       1
#define ICON_SMARROW    2

typedef struct menu_item
{
  char *text;
  int id;
  int icon;
  struct menu *sub_menu;
} menu_item_t;

typedef struct menu 
{
  struct 
  {
    struct 
    {
      float r, g, b;
    } border;      // Border
  
    struct 
    {
      float r, g, b;
    } in;       // Inside

    struct 
    {
      float r, g, b;
    } sel;     // Selection

    struct 
    {
      float r, g, b;
    } icon;     // Icon
  } col;

  struct 
  {
    float x, y;
    float w, h;    
    float z;   // Depth 
    float b;   // Border width
    float lh;  // Line space height
    float mw, mh;   // Upper-left margins
  } dim;

  struct 
  {
    menu_item_t *list;
    int cnt;
    int sel;
    int start;    // For scrolling menus
    int max;
    plx_font_t* fnt;
    plx_fcxt_t* fnt_cxt;   
  } items;
  
  struct 
  {
    int sub_open;   // Is sub menu open?
  } stat;

  struct 
  {
    int scroll;   
    int scroll_sm;  // scroll sub menus as well?
    int dim_text;
  } opt;
} menu_t;

typedef struct fileselector
{
  menu_t *menu;  // Menu used for display

  char *dir;

  int follow_dirs;    // Go in/out dirs

  struct
  {
    char *vector;     // Filenames
    int cnt;          // Count
  } fln;
} fileselector_t;

typedef struct requester
{
  menu_t choice_menu;   // Menu used for the choices
  menu_t msg_menu;   // Menu used for the message

  menu_item_t msg_menu_item;  // Menu item for message
} requester_t;

typedef struct key
{
  struct 
  {
    char *bot;
    char *top;
  } text;

  unsigned int id;

  float w; 
} key_t;

typedef struct key_row
{
  struct 
  {
    key_t *list;
    int cnt;
  } key;
  
  float x;
  float y;
  float h;
  float cw;

  struct
  {
    struct key_row *up;
    struct key_row *down;
    struct key_row *left;
    struct key_row *right;
  } dir;
} key_row_t;

typedef struct keyboard
{
  struct
  {
    key_row_t *list;
    int cnt;
  } row;
  
  struct
  {
    key_row_t *row;
    int index;
  } sel;

  struct 
  {
    plx_font_t* fnt;
    plx_fcxt_t* fnt_cxt;   
  } txt;

  struct 
  {
    float x, y;
    float w, h;    
    float z;  
    float b;   
  } dim;

  struct 
  {
    struct 
    {
      float r, g, b;
    } border;      // Border
    
    struct 
    {
      float r, g, b;
    } in;       // Inside
    
    struct 
    {
      float r, g, b;
    } sel;     // Selection   

    struct 
    {
      float r, g, b;
    } key;     // Selection   
  } col;
  
} keyboard_t;

typedef struct controller
{
  struct
  {
    cont_state_t last_st;
    int port;
  } cont;
 
  struct
  {
    int intval;
    int t_left;
  } rpt;

} controller_t;

void menu_create(menu_t *mn, float x, float y, float w, float h, float z, menu_item_t *items, int cnt, plx_font_t* fnt, plx_fcxt_t* fnt_cxt);
void menu_auto_sub_menu_pos(menu_t *mn);
void menu_draw_box_chain_pvr(menu_t *mn);
void menu_draw_text_chain_pvr(menu_t *mn);
int menu_operate(menu_t *mn, cont_state_t *cont_state);

void keyboard_create(keyboard_t *keybd, float x, float y, float w, float h, float z, 
		     plx_font_t* fnt, plx_fcxt_t* fcxt, key_row_t *row_list, int row_cnt);
int keyboard_operate(keyboard_t *keybd, cont_state_t *cont_state);
void keyboard_draw_box_pvr(keyboard_t *keybd);
void keyboard_draw_text_pvr(keyboard_t *keybd);

void fileselector_create(fileselector_t *fls, menu_t *menu, char *dir);
void fileselector_destroy(fileselector_t *fls);
int fileselector_operate(fileselector_t *fls, cont_state_t *cont_state, char **fln_out);

void requester_create(requester_t *req, float x, float y, float w, float h, float cw, float ch, float z, 
		      char *msg, menu_item_t *choice_items, int choice_cnt, 
		      plx_font_t* fnt, plx_fcxt_t* fnt_cxt);

void controller_create(controller_t *ctlr, int port, int intval);
bool controller_check_input(controller_t *ctlr);

void box_draw_pvr(float x, float y, float w, float h, float z, float b,
		  float ia, float ir, float ig, float ib, 
		  float ba, float br, float bg, float bb);
void rect_draw_pvr(float x1, float y1, float x2, float y2, float z, 
		   float a, float r, float g, float b);


#endif
