/* 
 * Dreamcast preferences menu 
 *
 * by Tolga ABACI, 2003
 */

#include "DCGUI.h"
#include <zlib/zlib.h>

#define MC_D8_DISK   1
#define MC_D9_DISK   2
#define MC_D10_DISK  3
#define MC_D11_DISK  4
#define MC_D8_TAPE   5
#define MC_D9_TAPE   6
#define MC_D10_TAPE  7
#define MC_D11_TAPE  8
#define MC_D8_EJECT  9
#define MC_D9_EJECT  10
#define MC_D10_EJECT 11
#define MC_D11_EJECT 12
#define MC_D8_AUTO   13
#define MC_D9_AUTO   14
#define MC_D10_AUTO  15
#define MC_D11_AUTO  16

#define MC_KEYBUI    22      
#define MC_PAL       23      
#define MC_NTSC      24
#define MC_DISPMODE  25
#define MC_BORDERS   26
#define MC_SPRITES   27
#define MC_COLLS     28
#define MC_SPDMETER  30
#define MC_LEDS      31
#define MC_FASTRESET 32
#define MC_FNCYRESET 33
#define MC_CIAIRQ    34
#define MC_SWAPJOY   35
#define MC_SOUND     36
#define MC_SIDFILTS  37
#define MC_SOFTRESET 38
#define MC_HARDRESET 39
#define MC_1541      40
#define MC_ABOUT     41
#define MC_NOSKIP    42      
#define MC_50SKIP    43      
#define MC_75SKIP    44
#define MC_LOADSTATE 45
#define MC_SAVESTATE 46
#define MC_QLOADSTAT 47
#define MC_QSAVESTAT 48
#define MC_DELSTATE  49
#define MC_KBB       51
#define MC_KBY       52
#define MC_KBL       53
#define MC_KBR       54
#define MC_OPTLDCD   56
#define MC_OPTLDVMU  57
#define MC_OPTSVVMU  58


static char last_state_fln[MAX_FN_LEN] = "";

static char* vmu_load_msg = "Choose VMU to load from...";
static char* vmu_save_msg = "Choose VMU to save into...";
static char* vmu_delete_msg = "Choose VMU to delete from...";
static char* vfl_load_msg = "Choose file to load...";
static char* vfl_delete_msg = "Choose file to delete...";

static char* buttons_text[] = { 
  "Choose key for button B...", 
  "Choose key for button Y...", 
  "Choose key for left trigger...",
  "Choose key for right trigger..."
};

static menu_t 
  main_menu, 
  drive_menu, 
  state_menu, 
  option_menu, 
  display_menu, 
  emulation_menu, 
  frameskip_menu, 
  misc_menu, 
  keybd_menu,
  reset_menu, 
  mode_menu, 
  df_menu,
  vmu_menu,
  msg_menu,
  sf_menu,
  d_box[4], 
  d_menu[4]; 

static fileselector_t drive_fls, vmu_fls, state_fls;

static menu_item_t msg_items[] = {
  { 0, 0, ICON_NONE, 0 }
};

static menu_item_t main_items[] = {
  { "Drives", 0, ICON_SMARROW, &drive_menu }, 
  { "States", 0, ICON_SMARROW, &state_menu },
  { "Options", 0, ICON_SMARROW, &option_menu },
  { "Reset", 0, ICON_SMARROW, &reset_menu},
  { "About...", MC_ABOUT, 0, 0 }
};

static menu_item_t mode_items[] = {
  { "PAL", MC_PAL, 0, 0 }, 
  { "NTSC", MC_NTSC, 0, 0 }
};

static menu_item_t state_items[] = {
  { "Load...", MC_LOADSTATE, 0, 0 },
  { "Save...", MC_SAVESTATE, 0, 0 },
  { "Quick Load", MC_QLOADSTAT, 0, 0 },
  { "Quick Save", MC_QSAVESTAT, 0, 0 },
  { "Delete...", MC_DELSTATE, 0, 0 }
};

static menu_item_t frameskip_items[] = {
  { "None", MC_NOSKIP, 0, 0 }, 
  { "50%", MC_50SKIP, 0, 0 },
  { "75%", MC_75SKIP, 0, 0 }
};

static menu_item_t drive_items[] = {
  { "8", 0, ICON_SMARROW, &d_menu[0] },
  { "9", 0, ICON_SMARROW, &d_menu[1] },
  { "10", 0, ICON_SMARROW, &d_menu[2] },
  { "11", 0, ICON_SMARROW, &d_menu[3] }
};

static menu_item_t d_menu_items[][4] = 
{
  {
    { "Insert Disk...", MC_D8_DISK, 0, 0 },
    { "Insert Tape...", MC_D8_TAPE, 0, 0 },
    { "Autostart", MC_D8_AUTO, 0, 0 },
    { "Remove Media", MC_D8_EJECT, 0, 0 }
  },
  {
    { "Insert Disk...", MC_D9_DISK, 0, 0 },
    { "Insert Tape...", MC_D9_TAPE, 0, 0 },
    { "Autostart", MC_D9_AUTO, 0, 0 },
    { "Remove Media", MC_D9_EJECT, 0, 0 }
  },
  {
    { "Insert Disk...", MC_D10_DISK, 0, 0 },
    { "Insert Tape...", MC_D10_TAPE, 0, 0 },
    { "Autostart", MC_D10_AUTO, 0, 0 }, 
    { "Remove Media", MC_D10_EJECT, 0, 0 }
  },
  {
    { "Insert Disk...", MC_D11_DISK, 0, 0 },
    { "Insert Tape...", MC_D11_TAPE, 0, 0 },
    { "Autostart", MC_D11_AUTO, 0, 0 },
    { "Remove Media", MC_D11_EJECT, 0, 0 }
  }
};

static menu_item_t keybd_items[] = 
{
  { "Button B...", MC_KBB, ICON_NONE, 0 },
  { "Button Y...", MC_KBY, ICON_NONE, 0 },
  { "L. Trigger...", MC_KBL, ICON_NONE, 0 },
  { "R. Trigger...", MC_KBR, ICON_NONE, 0 }
};

static menu_item_t d_box_items[][4] = 
{
  { { 0, 0, 0, 0 } },
  { { 0, 0, 0, 0 } },
  { { 0, 0, 0, 0 } },
  { { 0, 0, 0, 0 } }
};

static menu_item_t option_items[] = {
  { "Display", 0, ICON_SMARROW, &display_menu},
  { "Emulation", 0, ICON_SMARROW, &emulation_menu},
  { "Keyboard", 0, ICON_SMARROW, &keybd_menu},
  { "Misc", 0, ICON_SMARROW, &misc_menu},
  { "Load (CD)", MC_OPTLDCD, ICON_NONE, 0},
  { "Load (VMU)", MC_OPTLDVMU, ICON_NONE, 0},
  { "Save (VMU)", MC_OPTSVVMU, ICON_NONE, 0},
};

static menu_item_t display_items[] = {
  { "Mode", MC_DISPMODE, ICON_SMARROW, &mode_menu},
  { "Borders", MC_BORDERS, 0, 0},
  { "Speedometer", MC_SPDMETER, 0, 0 },
  { "Drive LEDs", MC_LEDS, 0, 0 }
};

static menu_item_t emulation_items[] = {
  { "Sprites", MC_SPRITES, 0, 0 },
  { "Collisions", MC_COLLS, 0, 0 },
  { "Skip Frames", 0, ICON_SMARROW, &frameskip_menu},
  { "Fast Reset", MC_FASTRESET, 0, 0},
  { "CIA IRQ Hack", MC_CIAIRQ, 0, 0},
  { "1541 Processor", MC_1541, 0, 0},
  { "Sound", MC_SOUND, 0, 0},
  { "SID Filters", MC_SIDFILTS, 0, 0}
};

static menu_item_t misc_items[] = {
  { "Swap Joysticks", MC_SWAPJOY, 0, 0},
  { "Keyboard UI", MC_KEYBUI, 0, 0},
  { "Fancy Reset", MC_FNCYRESET, 0, 0},
};

static menu_item_t reset_items[] = {
  { "Warm", MC_SOFTRESET, 0, 0 }, 
  { "Cold", MC_HARDRESET, 0, 0 },
};

static controller_t contr;

void about_box_draw_pvr()
{
  pvr_poly_hdr_t hdr;
  pvr_poly_cxt_t cxt;
  
  pvr_poly_cxt_col(&cxt, PVR_LIST_OP_POLY);
  cxt.gen.shading = PVR_SHADE_FLAT;
  pvr_poly_compile(&hdr, &cxt);
  pvr_prim(&hdr, sizeof(hdr));
  
  box_draw_pvr(120.0f, 180.0f, 400.0f, 220.0f, 11.0f, 2.0f, 
	       1.0f, 0.3f, 0.1f, 0.1f, 1.0f, 0.8f, 0.8f, 0.8f);
  box_draw_pvr(122.0f, 270.0f, 360.0f, 128.0f, 11.1f, 2.0f, 
	       1.0f, 0.4f, 0.1f, 0.1f, 1.0f, 0.4f, 0.1f, 0.1f);
  box_draw_pvr(218.0f, 182.0f, 300.0f, 88.0f, 11.1f, 2.0f, 
	       1.0f, 0.5f, 0.1f, 0.1f, 1.0f, 0.5f, 0.1f, 0.1f);
}

void about_text_draw_pvr(plx_fcxt_t *ss_fcxt)
{
  plx_fcxt_setsize(ss_fcxt, 24.0f);
  
  plx_fcxt_begin(ss_fcxt);
  plx_fcxt_setpos(ss_fcxt, 230.0f, 220.f, 12.0f);
  plx_fcxt_draw(ss_fcxt, "DreamFrodo v0.94");
  plx_fcxt_end(ss_fcxt);

  plx_fcxt_setsize(ss_fcxt, 18.0f);

  plx_fcxt_begin(ss_fcxt);
  plx_fcxt_setpos(ss_fcxt, 300.0f, 250.f, 12.0f);
  plx_fcxt_draw(ss_fcxt, "A C64 Emulator");
  plx_fcxt_end(ss_fcxt);

  plx_fcxt_begin(ss_fcxt);
  plx_fcxt_setpos(ss_fcxt, 140.0f, 310.f, 12.0f);
  plx_fcxt_draw(ss_fcxt, "Original (Frodo 4.1) by");
  plx_fcxt_end(ss_fcxt);

  plx_fcxt_begin(ss_fcxt);
  plx_fcxt_setpos(ss_fcxt, 300.0f, 330.f, 12.0f);
  plx_fcxt_draw(ss_fcxt, "Christian Bauer");
  plx_fcxt_end(ss_fcxt);

  plx_fcxt_begin(ss_fcxt);
  plx_fcxt_setpos(ss_fcxt, 140.0f, 360.f, 12.0f);
  plx_fcxt_draw(ss_fcxt, "Dreamcast Port by");
  plx_fcxt_end(ss_fcxt);

  plx_fcxt_begin(ss_fcxt);
  plx_fcxt_setpos(ss_fcxt, 300.0f, 380.f, 12.0f);
  plx_fcxt_draw(ss_fcxt, "Tolga Abaci");
  plx_fcxt_end(ss_fcxt);
}

char *find_last_slash(char *str)
{
  char *lsp = str;
  char *sp = str;
  
  while (*sp)
  {
    if (*sp == '/')
      lsp = sp;
    sp ++;
  }
  
  return lsp;
}

bool save_to_vmu(char *save_fln, bool verbose)
{
  char vmu_fln[MAX_FN_LEN];
  char *ls = find_last_slash(save_fln);
  *ls = 0;
  strcpy(vmu_fln, save_fln);
  *ls = '/'; 

  // First save to RAM, then compress, then copy into VMU.
  TheApp->TheC64->SaveSnapshot("/ram/save.tmp");
  
  gzFile gf = gzopen("/ram/save.gz", "wb9");
  file_t rf = fs_open("/ram/save.tmp", O_RDONLY);
  if (gf == NULL || rf == FILEHND_INVALID)
    goto error;

  stat_t vmu_stat;
  ssize_t gf_size;
  ssize_t count;
  int blocks;
  char buffer[512];
  
  while ((count = fs_read(rf, buffer, 512)))
    if (!gzwrite(gf, buffer, count))
      goto error;
  
  fs_close(rf);
  gzclose(gf);
    
  rf = fs_open("/ram/save.gz", O_RDONLY);
  fs_seek(rf, 0, SEEK_END);
  gf_size = fs_tell(rf);
  fs_close(rf);
  blocks = (gf_size / 512) + 1;

  if (fs_stat(vmu_fln, &vmu_stat) == -1)
    goto error;
  
  if (vmu_stat.size * 512 < gf_size)
  {
    char msg_fln[100];
                      
    sprintf(msg_fln, "Not enough free blocks (%d/%d)", vmu_stat.size, blocks);

    ShowRequester(msg_fln, "OK", NULL);

    fs_unlink("/ram/save.tmp");
    fs_unlink("/ram/save.gz");	      

    return false;
  }

  ShowMessage("Saving, do not remove VMU.");
  ShowMessage("Saving, do not remove VMU.");

  fs_copy("/ram/save.gz", save_fln);

  fs_unlink("/ram/save.tmp");
  fs_unlink("/ram/save.gz");	      
 
  if (verbose)
    ShowRequester("Save to VMU successful.", "OK", NULL);

  return true;  

  error:

  fs_unlink("/ram/save.tmp");
  fs_unlink("/ram/save.gz");	        

  ShowRequester("Error while saving to VMU.", "OK", NULL);
  
  return false;
}

bool load_from_vmu(char *state_fln, bool verbose)
{
  gzFile vf = gzopen(state_fln, "rb");
  file_t rf = fs_open("/ram/save.tmp", O_WRONLY);

  if (vf == NULL || rf == FILEHND_INVALID)
    goto error;
  
  unsigned int count;
  char buffer[1024];
  
  while ((count = gzread(vf, buffer, 1024)))
    if (fs_write(rf, buffer, count) != count)
      goto error;
  
  fs_close(rf);
  gzclose(vf);
  
  TheApp->TheC64->LoadSnapshot("/ram/save.tmp");
  
  fs_unlink("/ram/save.tmp");

  if (verbose)
    ShowRequester("Load from VMU successful.", "OK", NULL);

  return true;  

  error:

  fs_unlink("/ram/save.tmp");

  ShowRequester("Error while loading from VMU.", "OK", NULL);
  
  return false;

}

bool load_prefs_from_vmu(Prefs *prefs)
{
  char fln[MAX_FN_LEN];
  char *vmu = 0;

  file_t dir_hnd = fs_open("/vmu", O_DIR | O_RDONLY);
  while (dirent_t *direntry = fs_readdir(dir_hnd))
    if (strcmp(direntry->name, ".") && strcmp(direntry->name, ".."))
      vmu = direntry->name;

  if (vmu == 0)
    return false;

  sprintf(fln, "/vmu/%s/frodo.prf", vmu);
  
  file_t file_hnd = fs_open(fln, O_RDONLY);
  if (file_hnd == FILEHND_INVALID)
    return false; 
  fs_close(file_hnd);
  
  prefs->Load(fln);

  return true;
}

bool save_prefs_to_vmu(Prefs *prefs)
{
  char fln[MAX_FN_LEN];
  char *vmu = 0;

  file_t dir_hnd = fs_open("/vmu", O_DIR | O_RDONLY);
  while (dirent_t *direntry = fs_readdir(dir_hnd))
    if (strcmp(direntry->name, ".") && strcmp(direntry->name, ".."))
      vmu = direntry->name;

  if (vmu == 0)
    return false;

  sprintf(fln, "/vmu/%s/frodo.prf", vmu);
  
  prefs->Save(fln);

  return true;
}

static bool init = false;

#define STATE_MAIN_MENU 0
#define STATE_ABOUT_BOX 1
#define STATE_VMU_L_SEL 2
#define STATE_VMU_S_SEL 3
#define STATE_VMU_D_SEL 4
#define STATE_VFL_L_SEL 5
#define STATE_VFL_D_SEL 6
#define STATE_IMAGE_SEL 7
#define STATE_KEY_SEL   8

bool Prefs::ShowEditor(bool startup, char *prefs_name)
{
  bool quit = false;
  int df_active = 0;
  int ks_active = 0;
  int menu_state = STATE_MAIN_MENU;
  char *df_fln, *vmu_fln, *state_fln;

  plx_font_t *menu_fnt, *fls_fnt;
  plx_fcxt_t *menu_fcxt, *fls_fcxt;

  menu_fnt = TheApp->TheC64->TheDisplay->ss_fnt;
  menu_fcxt = TheApp->TheC64->TheDisplay->ss_fcxt;  
  fls_fnt = TheApp->TheC64->TheDisplay->fs_fnt;
  fls_fcxt = TheApp->TheC64->TheDisplay->fs_fcxt;  

  menu_create(&msg_menu, 80.0f, 360.0f, 480.0f, 80.0f, 10.0f, msg_items, 1, menu_fnt, menu_fcxt);
  msg_menu.dim.mw = 0.1f * 480.0f;
  msg_menu.dim.mh = 0.4f * 80.0f;
  msg_menu.col.in.r = 0.5f;
  msg_menu.col.in.g = 0.1f;
  msg_menu.col.in.b = 0.1f;
  msg_menu.items.sel = -1;

  menu_create(&main_menu, 30.0f, 50.0f, 150.f, 200.f, 10.0f, main_items, 5, menu_fnt, menu_fcxt);
  menu_create(&drive_menu, 0.0f, 0.0f, 80.f, 120.f, 11.0f, drive_items, 4, menu_fnt, menu_fcxt);
  for (int i = 0; i < 4; i ++)
    menu_create(&d_menu[i], 0.0f, 0.0f, 200.0f, 115.0f, 12.0f, d_menu_items[i], 4, menu_fnt, menu_fcxt);
  menu_create(&option_menu, 0.0f, 0.0f, 160.0f, 190.0f, 11.0f, option_items, 7, menu_fnt, menu_fcxt);
  menu_create(&state_menu, 0.0f, 0.0f, 160.0f, 140.0f, 11.0f, state_items, 5, menu_fnt, menu_fcxt);
  menu_create(&display_menu, 0.0f, 0.0f, 200.0f, 115.0f, 12.0f, display_items, 4, menu_fnt, menu_fcxt);
  menu_create(&emulation_menu, 0.0f, 0.0f, 200.0f, 220.0f, 12.0f, emulation_items, 8, menu_fnt, menu_fcxt);
  menu_create(&keybd_menu, 0.0f, 0.0f, 160.0f, 115.0f, 12.0f, keybd_items, 4,menu_fnt, menu_fcxt);
  menu_create(&frameskip_menu, 0.0f, 0.0f, 100.0f, 90.0f, 13.0f, frameskip_items, 3, menu_fnt, menu_fcxt);
  menu_create(&misc_menu, 0.0f, 0.0f, 200.f, 90.0f, 12.0f, misc_items, 3, menu_fnt, menu_fcxt);
  menu_create(&mode_menu, 0.0f, 0.0f, 100.f, 70.0f, 13.0f, mode_items, 2, menu_fnt, menu_fcxt);
  menu_create(&reset_menu, 0.0f, 0.0f, 100.f, 70.0f, 11.0f, reset_items, 2, menu_fnt, menu_fcxt);
  menu_create(&df_menu, 50.0f, 50.0f, 500.f, 350.f, 10.0f, main_items, 4, fls_fnt, fls_fcxt);
  menu_create(&vmu_menu, 50.0f, 50.0f, 200.f, 250.f, 10.0f, main_items, 4, fls_fnt, fls_fcxt);
  menu_create(&sf_menu, 300.0f, 50.0f, 200.f, 250.f, 10.0f, main_items, 4, fls_fnt, fls_fcxt);
  menu_auto_sub_menu_pos(&main_menu);

  if (!init)  // FIXME
  {
    init = true;
    char *dir = (char *) malloc((strlen("/") + 1) * sizeof(char));
    strcpy(dir, "/");
    fileselector_create(&drive_fls, &df_menu, dir);        

    for (int i = 0; i < 4; i ++)
    {
      menu_create(&d_box[i], d_menu[i].dim.x, d_menu[i].dim.y + d_menu[i].dim.h + 20.f, 320.0f, 36.0f, 12.0f, 
		  d_box_items[i], 1, fls_fnt, fls_fcxt);
      d_box[i].col.in.r = 0.5f;
      d_box[i].col.in.g = 0.1f;
      d_box[i].col.in.b = 0.1f;
      d_box[i].items.sel = -1;
      d_box[i].items.list[0].text = find_last_slash(DrivePath[i]) + 1;
    }

    controller_create(&contr, 0, 20);
  }

  display_items[1].icon = (ShowBorders ? ICON_TICK : ICON_NONE);
  display_items[2].icon = (ShowSpdMeter ? ICON_TICK : ICON_NONE);
  display_items[3].icon = (ShowLEDs ? ICON_TICK : ICON_NONE);

  emulation_items[0].icon = (SpritesOn ? ICON_TICK : ICON_NONE);
  emulation_items[1].icon = (SpriteCollisions ? ICON_TICK : ICON_NONE);
  emulation_items[3].icon = (FastReset ? ICON_TICK : ICON_NONE);
  emulation_items[4].icon = (CIAIRQHack ? ICON_TICK : ICON_NONE);
  emulation_items[5].icon = (Emul1541Proc ? ICON_TICK : ICON_NONE);
  emulation_items[6].icon = ((SIDType == SIDTYPE_DIGITAL) ? ICON_TICK : ICON_NONE);
  emulation_items[7].icon = (SIDFilters ? ICON_TICK : ICON_NONE);
   
  misc_items[0].icon = (JoystickSwap ? ICON_TICK : ICON_NONE);
  misc_items[1].icon = (ShowKeyboard ? ICON_TICK : ICON_NONE);
  misc_items[2].icon = (FancyReset ? ICON_TICK : ICON_NONE);

  mode_items[0].icon = (PALMode ? ICON_TICK : ICON_NONE);
  mode_items[1].icon = (!PALMode ? ICON_TICK : ICON_NONE);

  frameskip_items[0].icon = ((SkipFrames == 1) ? ICON_TICK : ICON_NONE);
  frameskip_items[1].icon = ((SkipFrames == 2) ? ICON_TICK : ICON_NONE);
  frameskip_items[2].icon = ((SkipFrames == 4) ? ICON_TICK : ICON_NONE);

  while (!quit)
  {
    pvr_wait_ready();
    //    vid_waitvbl();
    //    while (pvr_check_ready() == -1) { };
    pvr_scene_begin();

    pvr_list_begin(PVR_LIST_OP_POLY);

    TheApp->TheC64->TheDisplay->SendToPVR(PVR_PACK_COLOR(1.0f, 0.5f, 0.5f, 0.5f));

    switch(menu_state)
    {
    case STATE_ABOUT_BOX:
      about_box_draw_pvr();
      break;
    case STATE_VFL_L_SEL:
    case STATE_VFL_D_SEL:
      menu_draw_box_chain_pvr(&sf_menu);
    case STATE_VMU_L_SEL:
    case STATE_VMU_S_SEL:
    case STATE_VMU_D_SEL:
      menu_draw_box_chain_pvr(&vmu_menu);
      menu_draw_box_chain_pvr(&msg_menu);
      break;
    case STATE_KEY_SEL:
      keyboard_draw_box_pvr(&(TheApp->TheC64->TheDisplay->keybd));
      menu_draw_box_chain_pvr(&msg_menu);
      break;
    case STATE_IMAGE_SEL:
      menu_draw_box_chain_pvr(&df_menu);
      break;
    case STATE_MAIN_MENU:
      menu_draw_box_chain_pvr(&main_menu);
      if (drive_menu.stat.sub_open)
	  menu_draw_box_chain_pvr(&(d_box[drive_menu.items.sel]));
      break;
    }

    pvr_list_finish();
    pvr_list_begin(PVR_LIST_TR_POLY);

    switch(menu_state)
    {
    case STATE_ABOUT_BOX:
      about_text_draw_pvr(menu_fcxt);
      break;
    case STATE_VFL_L_SEL:
    case STATE_VFL_D_SEL:
      menu_draw_text_chain_pvr(&sf_menu);
    case STATE_VMU_L_SEL:
    case STATE_VMU_S_SEL:
    case STATE_VMU_D_SEL:
      menu_draw_text_chain_pvr(&vmu_menu);
      menu_draw_text_chain_pvr(&msg_menu);
      break;
    case STATE_IMAGE_SEL:
      menu_draw_text_chain_pvr(&df_menu);
      break;
    case STATE_KEY_SEL:
      keyboard_draw_text_pvr(&(TheApp->TheC64->TheDisplay->keybd));
      menu_draw_text_chain_pvr(&msg_menu);
      break;
    case STATE_MAIN_MENU:
      menu_draw_text_chain_pvr(&main_menu);
      if (drive_menu.stat.sub_open)
	menu_draw_text_chain_pvr(&(d_box[drive_menu.items.sel]));
      break;
    }

    pvr_list_finish();
    pvr_scene_finish();

    if (controller_check_input(&contr))
    {
      switch (menu_state)
      {
      case STATE_ABOUT_BOX:
	quit = (contr.cont.last_st.buttons & CONT_B);       
        break;
      case STATE_IMAGE_SEL:
	{
	  int input = MENU_NOACTION;

	  input = fileselector_operate(&drive_fls, &contr.cont.last_st, &df_fln);      // FIXME

	  switch (input)
	  {
	  case MENU_CLOSE:
	    menu_state = STATE_MAIN_MENU;
	    break;
	  case FLS_OK:
	    quit = true;
	    int drive_no = 0;
	    // FIXME: 256
	    if (df_active >= MC_D8_DISK && df_active <= MC_D11_DISK)
	    {
	      drive_no = df_active - MC_D8_DISK;
	      DriveType[drive_no] = DRVTYPE_D64;
	    }
	    else if (df_active >= MC_D8_TAPE && df_active <= MC_D11_TAPE)
	    {
	      drive_no = df_active - MC_D8_TAPE;
	      DriveType[drive_no] = DRVTYPE_T64;
	    }
	    strcpy(DrivePath[drive_no], df_fln);
	    d_box[drive_no].items.list[0].text = find_last_slash(DrivePath[drive_no]) + 1;
	    free(df_fln);
	    fileselector_destroy(&drive_fls);    
	  }	
	}
	break;
      case STATE_VMU_L_SEL:
      case STATE_VMU_D_SEL:
	{
	  int input = MENU_NOACTION;

	  input = fileselector_operate(&vmu_fls, &contr.cont.last_st, &vmu_fln);      // FIXME

	  switch (input)
	  {
	  case MENU_CLOSE:
	    menu_state = STATE_MAIN_MENU;
	    break;
	  case FLS_OK:
	    if (menu_state == STATE_VMU_L_SEL)
	    {
	      menu_state = STATE_VFL_L_SEL;	
	      msg_items[0].text = vfl_load_msg;
	    }
	    else 
	    {
	      menu_state = STATE_VFL_D_SEL;
	      msg_items[0].text = vfl_delete_msg;	
	    }
	    fileselector_create(&state_fls, &sf_menu, vmu_fln);
	    state_fls.follow_dirs = 0;
	    vmu_menu.stat.sub_open = 1;
	    break;
	  }	
	}
	break;
      case STATE_VMU_S_SEL:
	{
	  int input = MENU_NOACTION;

	  input = fileselector_operate(&vmu_fls, &contr.cont.last_st, &vmu_fln);      // FIXME

	  switch (input)
	  {
	  case MENU_CLOSE:
	    menu_state = STATE_MAIN_MENU;
	    break;
	  case FLS_OK:
	    char base_fln[8];
	    char save_fln[30];	    
	    char msg_text[60];
	    char *img_name = find_last_slash(DrivePath[0]) + 1;	   

	    for (int i = 0; i < 7; i ++)
	      base_fln[i] = img_name[i];
	    base_fln[7] = 0;
	    
	    char n = '1';
	    do 
	    {
	      sprintf(save_fln, "%s/%s%c.FSS", vmu_fln, base_fln, n);
	      if (n == '0')
		n = 'A';
	      else 
		n ++;
	    } while (fs_open(save_fln, O_RDONLY));

	    sprintf(msg_text, "Save to %s ?", save_fln);

	    if (ShowRequester(msg_text, "Yes", "No"))	    
	      if (save_to_vmu(save_fln, true))
		strcpy(last_state_fln, save_fln);
	   
	    free(vmu_fln);
	    quit = true;

	    break;
	  }	
	}
	break;
      case STATE_VFL_L_SEL:
	{
	  int input = MENU_NOACTION;

	  input = fileselector_operate(&state_fls, &contr.cont.last_st, &state_fln);      // FIXME

	  switch (input)
	  {
	  case MENU_CLOSE:
	    menu_state = STATE_VMU_L_SEL;
	    msg_items[0].text = vmu_load_msg;
	    vmu_menu.stat.sub_open = 0;
	    free(vmu_fln);
	    break;
	  case FLS_OK:
	    char msg_text[60];

	    quit = true;

	    sprintf(msg_text, "Load from %s ?", state_fln);

	    if (ShowRequester(msg_text, "Yes", "No"))	   
	    {
	      ShowMessage("Loading, do not remove VMU.");
	      ShowMessage("Loading, do not remove VMU.");

	      if (load_from_vmu(state_fln, true))
		strcpy(last_state_fln, state_fln);
	    }
	    
	    free(vmu_fln);
	    break;
	  }	
	}
	break;
      case STATE_VFL_D_SEL:
	{
	  int input = MENU_NOACTION;

	  input = fileselector_operate(&state_fls, &contr.cont.last_st, &state_fln);      // FIXME

	  switch (input)
	  {
	  case MENU_CLOSE:
	    menu_state = STATE_VMU_D_SEL;
	    msg_items[0].text = vmu_delete_msg;
	    vmu_menu.stat.sub_open = 0;
	    free(vmu_fln);
	    break;
	  case FLS_OK:
	    char msg_text[60];

	    quit = true;

	    sprintf(msg_text, "Delete %s ?", state_fln);

	    if (ShowRequester(msg_text, "Yes", "No"))	   
	    {
	      ShowMessage("Deleting, do not remove VMU.");
	      ShowMessage("Deleting, do not remove VMU.");

	      if (fs_unlink(state_fln) == -1)
		ShowRequester("Error while deleting from VMU.", "OK", NULL);
	      else 
		ShowRequester("Delete from VMU successful.", "OK", NULL);
	    }
	    
	    free(vmu_fln);
	    break;
	  }	
	}
	break;
      case STATE_KEY_SEL:
	{
	  int input = keyboard_operate(&(TheApp->TheC64->TheDisplay->keybd), &contr.cont.last_st);

	  if (input != KEY_NOACTION)
	  {
	    KeyMap[ks_active - MC_KBB] = input;
	    ks_active = 0;
	    menu_state = STATE_MAIN_MENU;
	    quit = true;	    	    
	  }
	  else if (contr.cont.last_st.buttons & CONT_B)
	  {
	    KeyMap[ks_active - MC_KBB] = -1;
	    quit = true;
	  }
	}
	break;
      case STATE_MAIN_MENU:
        {
	  int input = MENU_NOACTION;	

	  input = menu_operate(&main_menu, &contr.cont.last_st);
	  
	  switch (input)
	  {
	  case MENU_CLOSE:
	    quit = true;  
	    break;
	  case MC_D8_DISK:
	  case MC_D9_DISK:
	  case MC_D10_DISK:
	  case MC_D11_DISK:
	  case MC_D8_TAPE:
	  case MC_D9_TAPE:
	  case MC_D10_TAPE:
	  case MC_D11_TAPE:
	    df_active = input;
	    menu_state = STATE_IMAGE_SEL;
	    fileselector_create(&drive_fls, &df_menu, 0);    
	    break;
	  case MC_BORDERS:
	    ShowBorders = !ShowBorders;
	    display_items[1].icon = (ShowBorders ? ICON_TICK : ICON_NONE);
	    break;
	  case MC_SPRITES:
	    SpritesOn = !SpritesOn;
	    emulation_items[0].icon = (SpritesOn ? ICON_TICK : ICON_NONE);
	    break;
	  case MC_COLLS:
	    SpriteCollisions = !SpriteCollisions;      
	    emulation_items[1].icon = (SpriteCollisions ? ICON_TICK : ICON_NONE);
	    break;
	    break;
	  case MC_SPDMETER:
	    ShowSpdMeter = !ShowSpdMeter;
	    display_items[2].icon = (ShowSpdMeter ? ICON_TICK : ICON_NONE);
	    break;
	  case MC_LEDS:
	    ShowLEDs = !ShowLEDs;
	    display_items[3].icon = (ShowLEDs ? ICON_TICK : ICON_NONE);
	    break;
	  case MC_FASTRESET:
	    FastReset = !FastReset;
	    emulation_items[3].icon = (FastReset ? ICON_TICK : ICON_NONE);
	    break;
	  case MC_FNCYRESET:
	    FancyReset = !FancyReset;
	    misc_items[2].icon = (FancyReset ? ICON_TICK : ICON_NONE);
	    break;
	  case MC_CIAIRQ:
	    CIAIRQHack = !CIAIRQHack;
	    emulation_items[4].icon = (CIAIRQHack ? ICON_TICK : ICON_NONE);
	    break;
	  case MC_1541:
	    Emul1541Proc = !Emul1541Proc;
	    emulation_items[5].icon = (Emul1541Proc ? ICON_TICK : ICON_NONE);
	    break;
	  case MC_SWAPJOY:
	    JoystickSwap = !JoystickSwap;
	    misc_items[0].icon = (JoystickSwap ? ICON_TICK : ICON_NONE);
	    break;
	  case MC_KEYBUI:
	    ShowKeyboard = !ShowKeyboard;
	    misc_items[1].icon = (ShowKeyboard ? ICON_TICK : ICON_NONE);
	    break;
	  case MC_SOUND:
	    SIDType = ((SIDType == SIDTYPE_DIGITAL) ? SIDTYPE_NONE : SIDTYPE_DIGITAL);
	    emulation_items[6].icon = ((SIDType == SIDTYPE_DIGITAL) ? ICON_TICK : ICON_NONE);
	    break;
	  case MC_SIDFILTS:
	    SIDFilters = !SIDFilters;
	    emulation_items[7].icon = (SIDFilters ? ICON_TICK : ICON_NONE);
	    break;
	  case MC_SOFTRESET:
	    quit = true;
	    TheApp->TheC64->Reset();
	    break;
	  case MC_HARDRESET:
	    quit = true;
	    TheApp->TheC64->QuitThyself();
	    break;
	  case MC_D8_EJECT:
	  case MC_D9_EJECT:
	  case MC_D10_EJECT:
	  case MC_D11_EJECT:
	    quit = true;
	    strcpy(DrivePath[input - MC_D8_EJECT], "");
	    d_box[input - MC_D8_EJECT].items.list[0].text = DrivePath[input - MC_D8_EJECT];
	    break;
	  case MC_D8_AUTO:
	  case MC_D9_AUTO:
	  case MC_D10_AUTO:
	  case MC_D11_AUTO:
	    quit = true;
	    AutoStartDrive = input - MC_D8_AUTO;
	    break;
	  case MC_PAL:
	    PALMode = true;
	    mode_items[0].icon = ICON_TICK;
	    mode_items[1].icon = ICON_NONE;
	    break;
	  case MC_NTSC:
	    PALMode = false;
	    mode_items[1].icon = ICON_TICK;
	    mode_items[0].icon = ICON_NONE;
	    break;
	  case MC_NOSKIP:
	    SkipFrames = 1;
	    frameskip_items[0].icon = ICON_TICK;
	    frameskip_items[1].icon = frameskip_items[2].icon = ICON_NONE;
	    break;
	  case MC_50SKIP:
	    SkipFrames = 2;
	    frameskip_items[1].icon = ICON_TICK;
	    frameskip_items[0].icon = frameskip_items[2].icon = ICON_NONE;
	    break;
	  case MC_75SKIP:
	    SkipFrames = 4;
	    frameskip_items[2].icon = ICON_TICK;
	    frameskip_items[1].icon = frameskip_items[0].icon = ICON_NONE;
	    break;
	  case MC_ABOUT:
	    menu_state = STATE_ABOUT_BOX;
	    break;
	  case MC_LOADSTATE:
	    menu_state = STATE_VMU_L_SEL;
	    fileselector_create(&vmu_fls, &vmu_menu, "/vmu");
	    vmu_fls.follow_dirs = 0;
	    msg_items[0].text = vmu_load_msg;
	    break;
	  case MC_QLOADSTAT:
	    if (!strcmp(last_state_fln, ""))
	      ShowRequester("Please do a regular load/save first.", "OK", NULL);
	    else            
	    {
	      ShowMessage("Loading, do not remove VMU.");
	      ShowMessage("Loading, do not remove VMU.");

	      load_from_vmu(last_state_fln, false);
	    }		
	    quit = true;
	    break;
	  case MC_SAVESTATE:
	    menu_state = STATE_VMU_S_SEL;
	    fileselector_create(&vmu_fls, &vmu_menu, "/vmu");
	    vmu_fls.follow_dirs = 0;
	    msg_items[0].text = vmu_save_msg;
	    break;
	  case MC_QSAVESTAT:
	    if (!strcmp(last_state_fln, ""))
	      ShowRequester("Please do a regular load/save first.", "OK", NULL);
	    else 
	    {
	      ShowMessage("Saving, do not remove VMU.");
	      ShowMessage("Saving, do not remove VMU.");

	      save_to_vmu(last_state_fln, false);
	    }		
	    quit = true;
	    break;
	  case MC_DELSTATE:
	    menu_state = STATE_VMU_D_SEL;
	    fileselector_create(&vmu_fls, &vmu_menu, "/vmu");
	    vmu_fls.follow_dirs = 0;
	    msg_items[0].text = vmu_delete_msg;
	    break;
	  case MC_KBB:
	  case MC_KBY:
	  case MC_KBL:	  
	  case MC_KBR:
	    {
	      menu_state = STATE_KEY_SEL;
	      ks_active = input;	      	      
	      if (!KeyboardUp)
	      {
		menu_create(&msg_menu, 80.0f, 40.0f, 480.0f, 80.0f, 10.0f, msg_items, 1, menu_fnt, menu_fcxt);
		msg_menu.dim.mw = 0.1f * 480.0f;
		msg_menu.dim.mh = 0.4f * 80.0f;
		msg_menu.col.in.r = 0.5f;
		msg_menu.col.in.g = 0.1f;
		msg_menu.col.in.b = 0.1f;
		msg_menu.items.sel = -1;
	      }
	      msg_items[0].text = buttons_text[ks_active - MC_KBB];	      
	    }
	    break;	  
	  case MC_OPTLDCD:
	    Load("/cd/frodo.prefs");
	    quit = true;
	    break;
	  case MC_OPTLDVMU:
	    load_prefs_from_vmu(this);
	    quit = true;
	    break;
	  case MC_OPTSVVMU:
	    save_prefs_to_vmu(this);
	    quit = true;
	    break;
	  }
	}
      }
    }  
  }

  return true;
}
