/*
 *  SID_DC.i - 6581 emulation, Dreamcast specific stuff
 *
 *  Frodo (C) 1994-1997,2002 Christian Bauer
 *  Dreamcast stuff by Tolga ABACI
 */

#include <kos.h>
#include <dc/spu.h>
#include <dc/g2bus.h>
#include <dc/sound/aica_cmd_iface.h>
#include <dc/sound/sound.h>
#include <dc/sound/stream.h>

#include "C64.h"
#include "VIC.h"

/* taken from KOS' snd_stream.c */
#define SPU_RAM_BASE		0xa0800000

static int started;
static uint32 ram_buffer_size = 2 * SAMPLE_FREQ / 50;   // RAM Buffer size in samples
static uint32 spu_buffer_size = 4096;            // Size of buffer in SPU memory
static uint32 last_pos;

static uint32 last_time_sec, last_time_msec;

#define _sample int16

static _sample* snd_buf; /* scratch buffer to generate sound data */
static uint32 spu_ram_sch1;

void DigitalRenderer::init_sound(void)
{
  snd_init();
  started = 0;
  ram_buffer_size = (((ram_buffer_size >> 4) << 4) + 16);
  snd_buf = (_sample*) malloc(ram_buffer_size * sizeof(_sample));
  ready = true;
}

DigitalRenderer::~DigitalRenderer()
{
  snd_shutdown();
}

static inline unsigned int aica_getpos(int chan)
{
  return(g2_read_32(SPU_RAM_BASE + AICA_CHANNEL(chan) + offsetof(aica_channel_t, pos)));
}

void DigitalRenderer::EmulateLine(void)
{
  sample_buf[sample_in_ptr] = volume;
  sample_in_ptr = (sample_in_ptr + 1) % SAMPLE_BUF_SIZE;

  if (!started)
    Resume();

  uint32 time_sec, time_msec;
  timer_ms_gettime(&time_sec, &time_msec);

  uint32 time_elapsed = (time_sec - last_time_sec) * 1000 + (time_msec - last_time_msec);

  if (time_elapsed >= 5)
  {
    last_time_sec = time_sec;
    last_time_msec = time_msec;
    
    uint32 load_size = (SAMPLE_FREQ * time_elapsed) / 1000;

    if (load_size > ram_buffer_size)
      load_size = ram_buffer_size;
    else 
      load_size = ((load_size >> 4) << 4) + 16;    

    uint32 aica_pos = aica_getpos(0);
    if (last_pos < aica_pos && (last_pos + load_size) > aica_pos) 
      return;      // Better solution?

    calc_buffer(snd_buf, load_size * sizeof(_sample) );
    
    if (last_pos + load_size > spu_buffer_size)
    {
      int first_part = (spu_buffer_size - last_pos);

      spu_memload(spu_ram_sch1 + last_pos * sizeof(_sample), snd_buf, first_part * sizeof(_sample));
      spu_memload(spu_ram_sch1, snd_buf + first_part, (load_size - first_part) * sizeof(_sample));

      last_pos += load_size;
      last_pos -= spu_buffer_size;      
    }
    else
    {
      spu_memload(spu_ram_sch1 + last_pos * sizeof(_sample), snd_buf, load_size * sizeof(_sample));
      
      last_pos += load_size;      
    }  
  }


}

void DigitalRenderer::Pause(void)
{
  AICA_CMDSTR_CHANNEL(tmp, cmd, chan);
  
  /* Stop stream */
  /* Channel 0 */
  cmd->cmd = AICA_CMD_CHAN;
  cmd->timestamp = 0;
  cmd->size = AICA_CMDSTR_CHANNEL_SIZE;
  cmd->cmd_id = 0;
  chan->cmd = AICA_CH_CMD_STOP;
  snd_sh4_to_aica(tmp, cmd->size);

  started = 0;
}

void DigitalRenderer::Resume(void)
{
  AICA_CMDSTR_CHANNEL(tmp, cmd, chan);

  started = 1;

  spu_ram_sch1 = snd_mem_malloc(spu_buffer_size * sizeof(_sample));
  spu_memset(spu_ram_sch1, 0, spu_buffer_size * sizeof(_sample));

  last_pos = spu_buffer_size / 2;

  cmd->cmd = AICA_CMD_CHAN;
  cmd->timestamp = 0;
  cmd->size = AICA_CMDSTR_CHANNEL_SIZE;
  cmd->cmd_id = 0;
  chan->cmd = AICA_CH_CMD_START;
  chan->base = spu_ram_sch1;
  chan->type = AICA_SM_16BIT;
  chan->length = spu_buffer_size * sizeof(_sample);
  chan->loop = 1;
  chan->loopstart = 0;
  chan->loopend = spu_buffer_size;
  chan->freq = SAMPLE_FREQ;
  chan->vol = 240;
  chan->pan = 127;
  snd_sh4_to_aica(tmp, cmd->size);

  timer_ms_gettime(&last_time_sec, &last_time_msec);
}




